/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was jointly developed by                                 */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation  and Texas Instruments            */ 
/****************************************************************************/

/****************************************************************************/
/*     Texas Instruments Predictive Embedded Zerotree (PEZW) Image Codec    */
/*     Copyright 1996, 1997, 1998 Texas Instruments	      		    */
/****************************************************************************/

#include "PEZW_ac.h"

/* declarations for zerotree coding */
#ifndef ABS
#define ABS(x) (((x)<0)?-(x):(x))
#endif
#ifndef max
#define max(a, b)       ((a) > (b) ? (a) : (b))
#endif

/* symbols actuallry used for AC coding */
#define IZER   0             /* isolated zero */
#define IVAL   1             /* isolated valued coefficient */
#define ZTRZ   2             /* zero tree root: zero */
#define ZTRV   3             /* valued zerotree root */

/* for internal use only, not sent as zerotree symbols  */
#define DZ     4               /* descendant of a ZTRV */
#define DS     5               /* descendant of a ZTRS */
#define S      6              /* positions already marked significant with nonzero descendents */
#define Z      7       	/* zero (default when undetermined), only used at encoder */
#define IS     8              /* isolated positions already marked significant, only used at encoder */

#define POS    9              /* only used at decoder */
#define NEG    10             /* only used at decoder */
#define ZTRPOS 11             /* only used at decoder */
#define ZTRNEG 12             /* only used at decoder */

#define Maximum_Decomp_Levels   10
#define No_of_states_context0   9
#define No_of_states_context1   1
#define No_of_contexts No_of_states_context0*No_of_states_context1*Maximum_Decomp_Levels


/* SYMBOLS FOR INTERNAL USE */


#define MAXPASS 20		/* not used anymore, but still in decoder */
#define MAXSNRLAYERS 20		/* Maximum number of SNR layers */
#define MAXSIZE      5000000    /* maximum bitstream size handled */
#define MAXDCBUFFER   100000    /* size for the DC buffer */
#define MAXZTRBUFFER 1000000  /* size for maximum buffer for zerotree coding */

#define Code_value_bits 16

/* Models for arithmetic coding */
/* THE SET OF SYMBOLS THAT MAY BE ENCODED. */

#define No_of_chars 4		/* Number of character symbols */
/* ORIGINALLY No_of_chars was 256, change back for debugging ascii strings */
/* USING 6 for zerotrees */
#define EOF_symbol      No_of_chars     /* Index of EOF symbol */
#define CONTINUE_symbol (No_of_chars-1)  /* Index of continue symbol */

#define No_of_symbols (No_of_chars+1) /* Total number of symbols */

#define Max_frequency_TI 127

/* the different models for the zerotree quantization */
/* freq[0]:  IZER symbol        or   0
   freq[1]:  IVAL symbol       or   1
   freq[2]:  ZTRZ symbol
   freq[3]:  ZTRV symbol    or Continue_symbol 
   freq[4]:  EOF symbol   */

Ac_model dom_pass, dom_pass2, sub_pass, end_pass;

/* model for dominant pass */
static int freq_dom[No_of_symbols] = {1, 1, 1, 1, 0};

/* model specific for last level of dominant pass encoding */
/* no zerotree possible */
static int freq_dom2[No_of_symbols] = {1, 1, 0, 0, 0};

/* model for subordinate pass */
static int freq_sub[No_of_symbols] = {1, 1, 0, 0, 0};

/* model for sending the EOF symbol */
static int freq_end[No_of_symbols] = {0, 0, 0, 4, 1};


/* context models */

static int freq_dom_IZER[No_of_symbols] =   {1, 1, 0, 0, 0};     
static int freq_dom_ZTRZ[No_of_symbols] =   {1, 1, 1, 1, 0};     
static int freq_dom_ZTRS[No_of_symbols] =   {0, 1, 0, 1, 0};

/* model specific for last level of dominant pass encoding */
/* no zerotree possible */
static int freq_dom2_IZER[No_of_symbols] =  {1, 1, 0, 0, 0};

/* debugging parameters */
#define DEBUG_BS               0
#define DEBUG_BS_DETAIL        0
#define DEBUG_ZTR              0
#define DEBUG_ZTR_CONTEXTS     0
#define DEBUG_ZTR_DETAILS      0
#define DEBUG_ZTR_DETAILS_DEC  0
#define DEBUG_ZTR_DEC          0
#define DEBUG_ZTR_DEC_BS       0
#define DEBUG_ZTR_CONTEXTS_DEC 0
#define DEBUG_DEC_BS           0
