/*
 *	  ShapeEnCoding.c
 */
/* This code was obtained from Lehigh University/Vector Vision, Inc.. 
   It was originally developed by OKI. 
   Dr. Hongqiao Sun/Weiping integrated it to the Lehigh/VVI software. 
   
   The copyrights or credits should belong to the following people and companies:
   
   Zhixiong Wu and Yoshihiro Ueda of OKI,
   Hongqiao Sun of Vector Vision,
   Weiping Li of Lehigh University,
   
   Minor modifications was made by Shipeng Li to integrated it to Sarnoff's SOL
   software.
*/

#include <stdio.h>
#include <string.h>

#include "ShapeCodec.h"
#include "ShapeCommon.h"
#include "shape_def.h"
#include "ShapeUtil.h"
extern int ObjectWidth, ObjectHeight;
extern int STO_const_alpha;
extern unsigned char STO_const_alpha_value;
static BSS *ShapeBitstream=NULL;
static int ShapeBitstreamLength=0;
#define Error -1

static int
decide_CR(int x,int y,int blkn,int blkx, unsigned char **BAB_org,unsigned char **BAB_dwn,int change_CR_disable,int alphaTH,unsigned char **shape);
static int
ShapeHeaderEncode(int i,int j,int blkx,SBI *infor);
static int
ShapeContentEncode(int i,int j,int bsize,char **BAB,SBI *infor);

int
ShapeEnCoding(unsigned char *InShape, int object_width, int object_height, int alphaTH, int change_CR_disable)
{
    int	blks		= 4;
    int	blkn		= 1<<blks;
    int	blkx		= (object_width+15)/16;
    int	blky		= (object_height+15)/16;
    SBI	ShapeInf, *infor;

    unsigned char **BAB		= malloc_2d_char(blkn,blkn);
    unsigned char **BABdown	= malloc_2d_char(blkn,blkn);
    unsigned char **BAB_ext	= malloc_2d_char(blkn+4,blkn+4);
    unsigned char **shape;
    int	i, j, k, l, status;
    int	bsize;
    int	CR;
    int	bits = 0;

    ObjectWidth = object_width;
    ObjectHeight = object_height;

    shape = (unsigned char **)malloc(sizeof(unsigned char *)*object_height);
    if(shape == NULL) {
      fprintf(stderr,"Memory allocation failed\n");
      return Error;
    }
    for(i=0;i<object_height;i++) shape[i] = InShape + i*object_width;

    infor = &ShapeInf;
    infor->BitStream = ShapeBitstream = (BSS *)malloc(sizeof(BSS));
    if(infor->BitStream==NULL) {
      fprintf(stderr,"Memory allocation failed\n");
      return Error;
    }
    infor->BitStream->bs = (unsigned char *)malloc(sizeof(unsigned char)*object_width*object_height);
    if(infor->BitStream->bs==NULL) {
      fprintf(stderr,"Memory allocation failed\n");
      return Error;
    }
    /* clear the output buffer */
    memset(infor->BitStream->bs, (unsigned char )0, object_width*object_height);
    infor -> change_CR_disable	= change_CR_disable; 
    infor -> alphaTH			= alphaTH; 
    infor -> block_size			= blkn;
    infor -> block_num_hor		= blkx;
    infor -> block_num_ver		= blky;
    infor -> shape_mode = malloc_2d_int(blky, blkx);
    infor -> CR = malloc_2d_int(blky, blkx);
    infor -> ST = malloc_2d_int(blky, blkx);
    InitBitstream(0,infor->BitStream);
    PutBitstoStream(1, change_CR_disable,infor->BitStream);
    PutBitstoStream(1, STO_const_alpha?1:0,infor->BitStream);
    if(STO_const_alpha)
      PutBitstoStream(8, STO_const_alpha_value,infor->BitStream);
    for ( j=0; j<blky; j++ ) {
      for ( i=0; i<blkx; i++ ) {
	/* get BAB. */
    	for ( l=0; l<blkn; l++ ) {
	  for ( k=0; k<blkn; k++ ) {
            BAB[l][k] = ((((j<<blks)+l)<ObjectHeight)
			 && (((i<<blks)+k) <ObjectWidth))?
	      ( shape[(j<<blks)+l][(i<<blks)+k]!=0 ):0;
	  }
    	}
	/* check content of BAB. */
        status = decide_CR(i,j,blkn,blkx,BAB,BABdown,change_CR_disable,alphaTH,
                           shape);
        if ( status==ALL0 || status==ALL255 ) {
	  
			/* All 0 or 255, hence no CAE is needed.  */
            infor -> shape_mode[j][i] = status;
            infor -> CR        [j][i] = 1;
            infor -> ST        [j][i] = 0;

            if ( ShapeHeaderEncode(i,j,blkx,infor)==Error ) {
                fprintf(stderr,"\n CAE arithmetic coding Error !\n");
                return  Error;
            }

            for ( l=0; l<blkn; l++ ) {
            for ( k=0; k<blkn; k++ ) {
	      if((j<<blks)+l < ObjectHeight && (i<<blks)+k < ObjectWidth)
                shape[(j<<blks)+l][(i<<blks)+k] = ( status==ALL0 ? 0 : 1 );
            }
            }

        } else {

            infor -> shape_mode[j][i] = BORDER;
            infor -> CR        [j][i] = CR = ( 1<<(status-CR1_1) );

            if ( ShapeHeaderEncode(i,j,blkx,infor) == Error ) {
                fprintf(stderr,"\n CAE arithmetic coding Error !\n");
                return  Error;
            }

			/* Add surrounding pixels for CAE. */
            AddBorderToBAB(i,j,blkn,CR,blkx,BABdown,BAB_ext,shape);

            bsize = blkn/CR;
            if ( ShapeContentEncode(i,j,bsize,BAB_ext,infor) == Error ) {
                fprintf(stderr,"\n CAE arithmetic coding Error !\n");
                return  Error;
            }

            if ( CR!=1 ) {
				/* Reconstruct shape from BAB */
                UpSampling_Still(i,j,blkn,CR,blkx,BABdown,BAB,shape);
            }

            for ( l=0; l<blkn; l++ ) {
            for ( k=0; k<blkn; k++ ) {
	      if((j<<blks)+l < ObjectHeight && (i<<blks)+k < ObjectWidth)
                shape[(j<<blks)+l][(i<<blks)+k] = ( BAB[l][k]==0 ? 0 : 1 );
            }
            }

        }
    }
    }

    ShapeBitstreamLength=infor -> BitStream_Length	= infor -> BitStream -> cnt;

    free_2d_char ( BAB, blkn );
    free_2d_char ( BABdown, blkn );
    free_2d_char ( BAB_ext, blkn+4 );
    free_2d_int  (infor->shape_mode, blky); 
    free_2d_int  (infor->CR, blky); 
    free_2d_int  (infor->ST, blky); 
  /*   printf("%x\n", *(infor->BitStream->bs)); */
   
    free(shape);
    return ( infor->BitStream_Length );
}

/* Merge Shape Bitstream into main bit stream */
void
MergeShapeBitstream()
{
  if(ShapeBitstream==NULL) {
    fprintf(stderr,"ShapeBitStream Not Available\n");
    exit(1);
  }
  InitBitstream(0,ShapeBitstream);
  BitStreamMerge(ShapeBitstreamLength, ShapeBitstream);
  ByteAlignmentEnc();
/*   printf("%x\n", *(ShapeBitstream->bs)); */
  free(ShapeBitstream->bs);
  free(ShapeBitstream);
  ShapeBitstream = NULL;
}

static int
CheckBABstatus(int blkn,unsigned char **BAB1,unsigned char **BAB2,int alphaTH)
{
    int	i, j, k, l;
    int	SAD;
    int	all0   = 0;
    int	all255 = 0;

    for ( j=0; j<blkn; j+=4 ) {
    for ( i=0; i<blkn; i+=4 ) {

        SAD = 0;
        if ( BAB2==(unsigned char **)NULL ) {

            for ( l=0; l<4; l++ )
            for ( k=0; k<4; k++ )
                SAD += ( BAB1[j+l][i+k]!=0 );

            if ( 16 *     SAD  > alphaTH ) all0   = 1;
            if ( 16 * (16-SAD) > alphaTH ) all255 = 1;

            if ( all0==1 && all255==1 )
                return (BORDER);

        } else {

            for ( l=0; l<4; l++ )
            for ( k=0; k<4; k++ )
                SAD += ( BAB1[j+l][i+k]!=BAB2[j+l][i+k] );

            if ( 16 * SAD > alphaTH )
                return (BORDER);

        }
    }
    }
    if ( all0==0 ) return ( ALL0   );
    else           return ( ALL255 );
}

static int
CheckMB_Gray(int BABsize,unsigned char **BAB)
{
    int		i, j;
    int		all0   = 0;
    int		all255 = 0;

    for (j=0; j<BABsize; j++) {
    for (i=0; i<BABsize; i++) {

        if ( BAB [j][i] == 0 )	all255 = 1;
        else					all0   = 1;

        if ( all0 == 1 && all255 == 1 )	return (BORDER);
    }
    }
    if ( all0 == 0 ) return ( ALL0   );
    else             return ( ALL255 );
}

static int
decide_CR(int x,int y,int blkn,int blkx, unsigned char **BAB_org,unsigned char **BAB_dwn,int change_CR_disable,int alphaTH,unsigned char **shape)
{
    int		status;
    int		i, j;

	/* estimation of all_0 or all_255 or gray at 16x16 first. */
    if ( (status=CheckBABstatus(blkn,BAB_org,(unsigned char **)NULL,alphaTH))!=BORDER )
        return (status);

    if ( change_CR_disable == 0 ) {
        unsigned char	**BAB_up = malloc_2d_char ( blkn, blkn );

		/* estimation of CR = 1/4 conversion error. */
        DownSampling_Still(BAB_org,BAB_dwn,blkn/4,4);
    	UpSampling_Still(x,y,blkn,4,blkx,BAB_dwn,BAB_up,shape);

        if ( CheckBABstatus(blkn,BAB_org,BAB_up,alphaTH)!=BORDER ) {
            free(BAB_up);
            if ( (status=CheckMB_Gray(blkn/4,BAB_dwn))!=BORDER )
                return (status);
            else
                return (CR1_4 );
        }

		/* estimation of CR = 1/2 conversion error. */
        DownSampling_Still ( BAB_org, BAB_dwn, blkn/2, 2 );
        UpSampling_Still   ( x, y, blkn, 2, blkx, BAB_dwn, BAB_up, shape );

        if ( CheckBABstatus ( blkn, BAB_org, BAB_up, alphaTH ) != BORDER ) {
            free ( BAB_up );
            if ((status = CheckMB_Gray (blkn/2, BAB_dwn)) != BORDER )
                return (status);
            else
                return (CR1_2 );
        }
    }
    for ( j=0; j<blkn; j++ ) {
    for ( i=0; i<blkn; i++ ) {
        BAB_dwn[j][i] = BAB_org[j][i];
    }
    }
    return (CR1_1);
}

static int
ShapeContentEncode(int i,int j,int bsize,char **BAB,SBI *infor)
{
    int		k, l;
    int		dir, dirMin;
    int		len, lenMin = INT_MAX;
    int		context;
    char	target;
    int		ctx_max = 1023;
    BSS		*bitstream[2];
    ArCoder	coder;


    for ( dir=0; dir<2; dir++ ) {

        bitstream[dir]		  = (BSS *)malloc( sizeof( BSS ) );
        bitstream[dir] -> bs  = (unsigned char *) malloc(bsize*bsize*sizeof(unsigned char));
	memset( bitstream[dir] -> bs, (unsigned char)0, bsize*bsize);
        InitBitstream ( 1, bitstream[dir] );

/*        if ( dir == HOR ) { */
        if ( dir == 1 ) {

            StartArCoder_Still ( &coder );

            for ( l=0; l<bsize; l++ ) {
            for ( k=0; k<bsize; k++ ) {
                target  = BAB [l+2][k+2];
                context = ( BAB [l+2+(+0)][k+2+(-1)] << 0 ) +
			( BAB [l+2+(+0)][k+2+(-2)] << 1 ) +
			( BAB [l+2+(-1)][k+2+(+2)] << 2 ) +
			( BAB [l+2+(-1)][k+2+(+1)] << 3 ) +
			( BAB [l+2+(-1)][k+2+(+0)] << 4 ) +
			( BAB [l+2+(-1)][k+2+(-1)] << 5 ) +
			( BAB [l+2+(-1)][k+2+(-2)] << 6 ) +
			( BAB [l+2+(-2)][k+2+(+1)] << 7 ) +
			( BAB [l+2+(-2)][k+2+(+0)] << 8 ) +
			( BAB [l+2+(-2)][k+2+(-1)] << 9 );

                if ( context > ctx_max ) {
                    fprintf ( stderr, "\n Shape context Error !\n" );
                    return  Error;
                }
                ArCodeSymbol_Still( &coder, bitstream[dir], target, intra_prob[context] );	
            }
            }
            StopArCoder_Still( &coder, bitstream[dir] );

            if ( (len  = bitstream[dir] -> cnt) < lenMin ) {
                lenMin = len;
                dirMin = dir;
            }

        } else {

            StartArCoder_Still ( &coder );

            for ( k=0; k<bsize; k++ ) {
            for ( l=0; l<bsize; l++ ) {
                target  = BAB [l+2][k+2];
                context = ( BAB [l+2+(-1)][k+2+(+0)] << 0 ) +
			( BAB [l+2+(-2)][k+2+(+0)] << 1 ) +
			( BAB [l+2+(+2)][k+2+(-1)] << 2 ) +
			( BAB [l+2+(+1)][k+2+(-1)] << 3 ) +
			( BAB [l+2+(+0)][k+2+(-1)] << 4 ) +
			( BAB [l+2+(-1)][k+2+(-1)] << 5 ) +
			( BAB [l+2+(-2)][k+2+(-1)] << 6 ) +
			( BAB [l+2+(+1)][k+2+(-2)] << 7 ) +
			( BAB [l+2+(+0)][k+2+(-2)] << 8 ) +
			( BAB [l+2+(-1)][k+2+(-2)] << 9 );

                if ( context > ctx_max ) {
                    fprintf ( stderr, "\n Shape context Error !\n" );
                    return  Error;
                }
                ArCodeSymbol_Still( &coder, bitstream[dir], target, intra_prob[context] );	

            }
            }
            StopArCoder_Still( &coder, bitstream[dir] );

            if ( (len  = bitstream[dir] -> cnt) < lenMin ) {
                lenMin = len;
                dirMin = dir;
            }

        }

    }

    infor -> ST[j][i] = dirMin;
	
    PutBitstoStream ( LST, (unsigned int)dirMin, infor -> BitStream );

    InitBitstream ( 0, bitstream[dirMin] );

    BitStreamCopy ( lenMin, bitstream[dirMin], infor -> BitStream );

    for ( dir=0; dir<2; dir++ ) {
        free ( bitstream[dir] -> bs );
        free ( bitstream[dir] );
    }

    return ( 0 );
}

static int
ShapeHeaderEncode(int i,int j,int blkx,SBI *infor)
{
    int	mode		= infor -> shape_mode[j][i];
    int	dis_CR		= infor -> change_CR_disable;
    BSS	*bitstream	= infor -> BitStream;

    int	UL = ( (i == 0      || j == 0) ? 0 : infor -> shape_mode[j-1][i-1] );
    int	UR = ( (i == blkx-1 || j == 0) ? 0 : infor -> shape_mode[j-1][i+1] );
    int	U  = ( (j == 0) ? 0 : infor -> shape_mode[j-1][i] );
    int	L  = ( (i == 0) ? 0 : infor -> shape_mode[j][i-1] );
    int		index	= ( 27 * UL + 9 * U + 3 * UR + L );
    int		bits;
    unsigned int	code;

    bits = LMMR_first_shape_code_I [index][mode];
    code = CMMR_first_shape_code_I [index][mode];

    PutBitstoStream ( bits, code, bitstream );

    if ( mode == BORDER ) {
        if ( dis_CR == 0 ) { 
            bits = LCR [infor -> CR[j][i]];
            code = CCR [infor -> CR[j][i]];

            PutBitstoStream ( bits, code, bitstream );
        }
    }

    return ( 0 );
}

