/* $Id: ac.h,v 1.6 1998/02/20 21:14:35 hjlee Exp $ */

/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

/************************************************************/
/*  Filename: ac.h                                          */
/*                                                          */
/*  Descriptions:                                           */
/*    This is the header file of ac.c containing arithmetic */
/*    coding routines.                                      */
/*                                                          */
/************************************************************/

#ifndef AC_HEADER
#define AC_HEADER

#include <stdio.h>

#define MAX_BUFFER 1024

typedef struct {
  FILE *fp;
  long low;
  long high;
  long fbits;
  int buffer;
  int bits_to_go;
  long total_bits;
  unsigned char *bitstream;
  long  bitstream_len;
} ac_encoder;

typedef struct {
  FILE *fp;
  long value;
  long low;
  long high;
  int buffer;
  int bits_to_go;
  unsigned char *bitstream;
  long total_bits;
  long  bitstream_ptr;
  int  dc_band;
} ac_decoder;

typedef struct {
  int nsym;   /* the number of symbols in the symbol set to be encoded */
  int adapt;  /* 0 => non-adaptive, !0 => adaptive */
  int order;    /* AC_ORDER_<ZER0, FIRST, or MIX> */

  /* for zeroth and mixed order models only */
  unsigned short *freq_0;
  unsigned short *cfreq_0;

  /* for first order and mixed order models only */
  unsigned short *freq_1;
  unsigned short *cfreq_1;
  int freq1Q;
  int numFreq1Slots;
  int context;   /* Current context in first and mixed models: 
		    if context < 0 then lastSym is used to figure out context
		    otherwise context should be in
		    [0,1,...,numFreq1Slots-1].
		    Note: if context >= 0 then it must be passed as last
		    argument in ac_encode_symbol and ac_decode_symbol 
		    if context < 0 the nothing need be passed (variable 
		    argument list) or if passed it's ignored. */
  int lastSym; /* Last symbol encoded: If context<0 then this is used to 
		  figure out context in first and mixed models. This is
		  done by the mapping: context=lastSym*numFreq1Slots/nsym */

  /* for mixed order model only */
  int firstOrderFreqInc;

} ac_model;


/* Values order field in ac_model_init function and ac_model structure */
#define AC_ORDER_ZERO  0
#define AC_ORDER_FIRST 1
#define AC_ORDER_MIX   99


int min_predict;


/* function prototypes */
/* void mput_bistream(unsigned char val, int nbits, ac_encoder *ace); */
void mput_bistream(int val, int nbits, ac_encoder *ace);
/* not currently used - also "un if 0" in ac.c if needed */
#if 0
long ac_encoder_bits (ac_encoder *ace);
#endif

void ac_encoder_init (ac_encoder *ace, const char *fn, 
		      unsigned char *bitstream);
int ac_encoder_done (ac_encoder *ace,unsigned char **bitbuffer);
int ac_encode_symbol (ac_encoder *ace, ac_model *acm, int sym, ...);


int mget_bitstream(int nbits, ac_decoder *acd);

void ac_decoder_open (ac_decoder *acd, const char *fn);
void ac_decoder_init (ac_decoder *acd, const char *fn, 
		      unsigned char *bitstream);
void ac_decoder_done (ac_decoder *acd);
int ac_decode_symbol (ac_decoder *acd, ac_model *acm, ...);

void ac_model_init (ac_model *acm, int nsym, int *ifreq, int adapt, int order,
		    int context, int firstOrderSlots, int firstOrderInc);
void ac_model_done (ac_model *acm);

/* not currently used - also "un if 0" in ac.c if needed */
#if 0
void ac_model_save (ac_model *acm, int *ifreq);
#endif

#endif
