/* $Id: ztscan_dec.c,v 1.41 1998/02/20 21:14:39 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

/************************************************************/
/*  Filename: ztscan_dec.c                                  */
/*  Author: Bing-Bing CHai                                  */
/*  Date: Dec. 17, 1997                                     */
/*                                                          */
/*  Descriptions:                                           */
/*    This file contains the routines that performs         */
/*    zero tree scanning and entropy decoding.              */
/*                                                          */
/************************************************************/

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <ctype.h>
#include <string.h>
#include <math.h>

#include "Utils.h"

#include "startcode.h"
#include "dataStruct.h"
#include "states.h"
#include "globals.h"

#include "ac.h"
#include "bitpack.h"


static COEFFINFO **coeffinfo;
static color;

/* definations related to arithmetic coding */
#define NUMCHAR_TYPE 4
#define ADAPT 1
#define PROB_MOD_ORDER AC_ORDER_ZERO
#define NO_CONTEXT -1
#define valSlotMax 10
#define valInc 15

/* arithmetic models related variables, refer to syntax for meaning */
static ac_decoder acd;
static ac_model acm_type, acm_vz, acm_vnz, acm_res, acm_root,acm_sign;

static int height,width;

#ifdef _CHECK_HI_LO_
FILE *acfp;
#endif



/* functions defined */
void wavelet_dc_decode(int c);
static void callc_decode();
void wavelet_higher_bands_decode_SQ(int c);
static void cachb_decode_SQ(SNR_IMAGE *snr_img);
static void decode_pixel_SQ(int h,int w);
static void  mag_sign_decode_SQ(int h,int w);
void wavelet_higher_bands_decode_MQ(SNR_IMAGE *snr_image, int c);
static void cachb_decode_MQ(SNR_IMAGE *snr_image);
void clear_ZTR_D();
static void decode_pixel_MQ(int h,int w);
static void mark_ZTR_D(int h,int w);
static void mag_sign_decode_MQ(int h,int w);


/* The following functions assume that mzte_codec is a global variable */
extern WVT_CODEC mzte_codec;

/******************************************************************/
/****************************  DC  ********************************/
/******************************************************************/

/*******************************************************/
/**************  Inverse DC Prediction  ****************/
/*******************************************************/

/********************************************************
  Function Name
  -------------
  static DATA iDC_pred_pix(int i, int j)

  Arguments
  ---------
  int i, int j: Index of wavelet coefficient (row, col)
  
  Description
  -----------
  Inverse DPCM prediction for a DC coefficient, refer 
  to syntax for algorithm. 

  Functions Called
  ----------------
  None.

  Return Value
  ------------
    inverse prediction for coeffinfo[i][j].quantized_value
********************************************************/ 
static DATA  iDC_pred_pix(int i, int j)
{
  DATA a,b,c;

  if(i==0){
    if(j==0)
      return(0);
    else
      return(coeffinfo[i][j-1].quantized_value);
  }
  else if(j==0)
    return(coeffinfo[i-1][j].quantized_value);

  a=coeffinfo[i][j-1].quantized_value;
  b=coeffinfo[i-1][j-1].quantized_value;
  c=coeffinfo[i-1][j].quantized_value;
  if(abs(a-b) < abs(b-c))
    return(c);
  else
    return(a);
}



/*****************************************************
  Function Name
  -------------
  void iDC_predict()

  Arguments
  ---------
  None
  
  Description
  -----------
  control program for inverse DC prediction

  Functions Called
  ----------------
  iDC_pred_pix(i,j).

  Return Value
  ------------
  none
******************************************************/
static void iDC_predict(int color)
{
  int i,j,dc_h,dc_w,offset_dc;

  dc_h=mzte_codec.dcHeight;
  dc_w=mzte_codec.dcWidth;
 
  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;
  offset_dc=mzte_codec.offset_dc;

  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++)
      coeffinfo[i][j].quantized_value += offset_dc;

  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++)
      coeffinfo[i][j].quantized_value+=iDC_pred_pix(i,j);
}




/********************************************************
  Function Name
  -------------
  void wavelet_dc_decode(int c)


  Arguments
  ---------
  int c - color component.
  
  Description
  -----------
  Control program for decode DC information for one 
  color component.

  Functions Called
  ----------------
  None.
  iDC_predict()
  get_param()
  cacll_decode()
  
  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_dc_decode(int c)
{
  noteDetail("Decoding DC (wavelet_dc_decode)....");
  color=c;

  mzte_codec.mean[color] = get_X_bits(8);
  /* mzte_codec.Qdc[color]  = get_X_bits(8); */
  mzte_codec.Qdc[color]  = get_param(7);

  mzte_codec.offset_dc=-get_param(7);
  /* mzte_codec.max_dc=get_param(7); */
  mzte_codec.max_dc=get_param(7)-mzte_codec.offset_dc; /* hjlee */

#ifdef _CHECK_HI_LO_
  acfp=fopen("decode_hl.dat","w");
#endif

  callc_decode();
  iDC_predict(color);
  noteDetail("Completed decoding DC.");

#ifdef _CHECK_HI_LO_
  fclose(acfp);
#endif


}



/********************************************************
  Function Name
  -------------
  static void cacll_decode()

  Arguments
  ---------
  None.

  
  Description
  -----------
  Decode DC information for one color component.

  Functions Called
  ----------------
  ac_decoder_init()
  ac_model_init()
  ac_decode_symbol()
  ac_model_done()
  ac_decoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void callc_decode()
{
  int dc_h, dc_w,i,j,offset=mzte_codec.offset_dc;

  dc_w=mzte_codec.dcWidth;
  dc_h=mzte_codec.dcHeight;

  /* init arithmetic model */
  /* ac_decoder_open(acd,NULL); */
  ac_decoder_init(&acd,NULL,NULL);
  ac_model_init(&acm_vz,mzte_codec.max_dc+1,
               NULL,ADAPT,PROB_MOD_ORDER,NO_CONTEXT,
               MIN(mzte_codec.max_dc+1,
                   valSlotMax),valInc);

  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;
  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++){

#ifdef _CHECK_HI_LO_
      fprintf(acfp,"%3d, %3d: %7d %7d\n",i,j,acd.high,acd.low);
#endif

      coeffinfo[i][j].quantized_value=ac_decode_symbol(&acd,&acm_vz);
    }

  /* close arithmetic coder */
  ac_model_done(&acm_vz);
  ac_decoder_done(&acd);
}



/*********************************************************************/
/*****************************  AC  **********************************/
/*************************  Single quant  ****************************/
/*********************************************************************/

/********************************************************
  Function Name
  -------------
  void wavelet_higher_bands_decode_SQ(int c)

  Arguments
  ---------
  int c - color component.
  
  Description
  -----------
  Control program for decoding AC information for one 
  color component. Single quant mode.

  Functions Called
  ----------------
  get_param()
  cachb_decode_SQ()

  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_higher_bands_decode_SQ(int c)
{
  SNR_IMAGE *snr_image;
  int marker_bit;

  noteDetail("Decoding AC band (wavelet_higher_bands_decode_SQ)....");
  color = c;
  snr_image=&(mzte_codec.SPlayer[color].SNRlayer.snr_image);

  /* header info */
  snr_image->root_max=get_param(7);
  marker_bit = get_X_bits(1);
  snr_image->valz_max=get_param(7);
  snr_image->valnz_max=get_param(7);

  cachb_decode_SQ(snr_image);
  noteDetail("Completed decoding AC band.");
}



/********************************************************
  Function Name
  -------------
  static void cachb_decode_SQ(SNR_IMAGE *snr_img)

  Arguments
  ---------
  SNR_IMAGE *snr_img - structure containing one color
                       component of an SNR image
  Description
  -----------
  Decode AC information for one color component. 
  Single quant mode, tree-depth scan

  Functions Called
  ----------------
  ac_decoder_init()
  ac_model_init()
  decode_pixel_SQ()
  ac_model_done()
  ac_decoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void cachb_decode_SQ(SNR_IMAGE *snr_img)
{
  int h,w,dc_h,dc_w,dc_h2,dc_w2;

  dc_h=mzte_codec.dcHeight;
  dc_w=mzte_codec.dcWidth;
  dc_h2=dc_h<<1;
  dc_w2=dc_w<<1;
  height=mzte_codec.SPlayer[color].height;
  width=mzte_codec.SPlayer[color].width;

  /* init arithmetic coder */
  ac_decoder_init(&acd,NULL,NULL);
  ac_model_init(&acm_type,NUMCHAR_TYPE,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(NUMCHAR_TYPE,valSlotMax),valInc);
  if (snr_img->root_max)
    ac_model_init(&acm_root,snr_img->root_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_img->root_max,valSlotMax),valInc);
  ac_model_init(&acm_vz,snr_img->valz_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_img->valz_max+1,valSlotMax),valInc);
  if (snr_img->valnz_max)
    ac_model_init(&acm_vnz,snr_img->valnz_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_img->valnz_max,valSlotMax),valInc);
  ac_model_init(&acm_sign,2,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(2,valSlotMax),valInc);

  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;

  /* LH */
  for(h=0;h<dc_h;h++)
    for(w=dc_w;w<dc_w2;w++){
      decode_pixel_SQ(h,w);           /* LH */
      decode_pixel_SQ(h+dc_h,w-dc_w); /* HL */
      decode_pixel_SQ(h+dc_h,w);      /* HH */
    }

  /* close arithmetic coder */
  ac_model_done(&acm_type);
  if (snr_img->root_max)
    ac_model_done(&acm_root);
  ac_model_done(&acm_vz);
  if (snr_img->valnz_max)
    ac_model_done(&acm_vnz);
  ac_model_done(&acm_sign);
  ac_decoder_done(&acd);
}



/********************************************************
  Function Name
  -------------
  static void decode_pixel_SQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel in height and width
  
  Description
  -----------
  Decoding the type and/or value of a coefficient, a
  recursive function.

  Functions Called
  ----------------
  mag_sign_decode_SQ()
  ac_decode_symbol()
  decode_pixel_SQ()

  Return Value
  ------------
  None.

********************************************************/ 
static void decode_pixel_SQ(int h,int w)
{
  int i,j;
  UNIT8 zt_type;

  /* decode leave coefficients, value only */

  
  if(coeffinfo[h][w].state == S_LEAF){ /* zero value. no sign */
#ifdef _SHAPE_
    if(coeffinfo[h][w].mask==1)
#endif      
      mag_sign_decode_SQ(h,w);
#ifdef _SHAPE_
    else 
      coeffinfo[h][w].quantized_value = 0;
#endif      
#ifdef _WRITE_COEFF_MAP_
    if (coeffinfo[h][w].state == S_LEAF)
      coeffinfo[h][w].type = (coeffinfo[h][w].quantized_value) ? VLEAF : ZLEAF;
#endif
      
      return;
  }
  
  
  /* decode zero tree symbols */
#ifdef _SHAPE_
  if(coeffinfo[h][w].mask==1)
#endif
  coeffinfo[h][w].type=zt_type=ac_decode_symbol(&acd,&acm_type);      
#ifdef _SHAPE_
  else
    coeffinfo[h][w].type=zt_type = IZ;
#endif
  /* code magnitude and sign */
  switch(zt_type){
    case IZ :
      break;
    case VZTR:
     mag_sign_decode_SQ(h,w);
    case ZTR:
#ifdef _WRITE_COEFF_MAP_
      mark_ZTR_D(h,w);  /* not necessary, only for checking purpose */
#endif _WRITE_COEFF_MAP_
      return;
    case VAL:
      mag_sign_decode_SQ(h,w);
      break;
    default: 
      errorHandler("Invalid zerotree symbol in single quant decode");
  }
  

  /* scan children */
  i=h<<1; j=w<<1;
  decode_pixel_SQ(i,j);
  decode_pixel_SQ(i,j+1);
  decode_pixel_SQ(i+1,j);
  decode_pixel_SQ(i+1,j+1);
}


     
/********************************************************
  Function Name
  -------------
  static void  mag_sign_decode_SQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  Description
  -----------
  Decode the value of a coefficient.

  Functions Called
  ----------------
  ac_decode_symbol()

  Return Value
  ------------
  None.

********************************************************/ 
static void  mag_sign_decode_SQ(int h,int w)
{
  int value,v_sign;

  switch(stateToProbModel[coeffinfo[h][w].state]){    
    case ACM_ROOT:
      value=ac_decode_symbol(&acd,&acm_root)+1;
      break;
    case ACM_VALNZ:
      value=ac_decode_symbol(&acd,&acm_vnz)+1;
      break;
    case ACM_VALZ:  /* leaf */
      if((value=ac_decode_symbol(&acd,&acm_vz))==0){
        coeffinfo[h][w].quantized_value=value;
        return;
      }
      break;
    default:
      errorHandler("not a valid magnitude type in encode_pixel_SQ");
  }
  v_sign=ac_decode_symbol(&acd,&acm_sign);
  coeffinfo[h][w].quantized_value=(v_sign==0) ? value : -value;
}



/*********************************************************************/
/******************************  AC  *********************************/
/**************************  Multi quant  ****************************/
/*********************************************************************/

/********************************************************
  Function Name
  -------------
  void wavelet_higher_bands_decode_MQ(SNR_IMAGE *snr_image,int c)

  Arguments
  ---------
  SNR_IMAGE *snr_image - structure containing one color
                         component of an SNR image
  int c - color component.
  
  Description
  -----------
  Control program for decoding AC information for one 
  color component. Multi quant mode.

  Functions Called
  ----------------
  get_param()
  cachb_decode_MQ()

  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_higher_bands_decode_MQ(SNR_IMAGE *snr_image, int c)
{
  int marker_bit;

  noteDetail("Decoding AC band (wavelet_higher_bands_decode_MQ)....");
  /* header info */
  snr_image->root_max=get_param(7);
  marker_bit = get_X_bits(1);
  snr_image->valz_max=get_param(7);
  snr_image->valnz_max=get_param(7);
  color=c;
  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;

  noteDebug("root=%d, valz=%d, valnz=%d",snr_image->root_max,
	     snr_image->valz_max,snr_image->valnz_max);

  cachb_decode_MQ(snr_image);
}




/********************************************************
  Function Name
  -------------
  static void cachb_decode_MQ(SNR_IMAGE *snr_img)

  Arguments
  ---------
  SNR_IMAGE *snr_img - structure containing one color
                       component of an SNR image 
  Description
  -----------
  Decode AC information for one color component. 
  Multiple quant, bandwise scan.

  Functions Called
  ----------------
  updateResidMaxAndAssignSkips
  ac_decoder_init()
  ac_model_init()
  clear_ZTR_D();
  decode_pixel_MQ()
  ac_model_done()
  ac_decoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void cachb_decode_MQ(SNR_IMAGE *snr_image)
{
  int h,w,ac_h,ac_w,ac_h2,ac_w2;

  /* ac_h, ac_w init */
  ac_h=mzte_codec.dcHeight;
  ac_w=mzte_codec.dcWidth;
  ac_h2=ac_h<<1;
  ac_w2=ac_w<<1;
  height=mzte_codec.SPlayer[color].height;
  width=mzte_codec.SPlayer[color].width;

  /* init arithmetic coder */
  ac_decoder_init(&acd,NULL,NULL);

  ac_model_init(&acm_type,NUMCHAR_TYPE,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(NUMCHAR_TYPE,valSlotMax),valInc);
  if (snr_image->root_max)
    ac_model_init(&acm_root,snr_image->root_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_image->root_max,valSlotMax),valInc);
  ac_model_init(&acm_vz,snr_image->valz_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_image->valz_max+1,valSlotMax),valInc);
  if (snr_image->valnz_max)
    ac_model_init(&acm_vnz,snr_image->valnz_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_image->valnz_max,valSlotMax),valInc);
  ac_model_init(&acm_res,snr_image->residual_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_image->residual_max+1,valSlotMax),valInc);
  ac_model_init(&acm_sign,2,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(2,valSlotMax),valInc);

  /* clear the ZTR_D type from the previous pass */
  clear_ZTR_D();

  /* scan each coefficients in the spatial layer */
  while(ac_h2<=height && ac_w2<=width){
    /* LH */
    for(h=0;h<ac_h;h++)
      for(w=ac_w;w<ac_w2;w++)
        decode_pixel_MQ(h,w);

    /* HL */
    for(h=ac_h;h<ac_h2;h++)
      for(w=0;w<ac_w;w++)
        decode_pixel_MQ(h,w);

    /* HH */
    for(h=ac_h;h<ac_h2;h++)
      for(w=ac_w;w<ac_w2;w++)
        decode_pixel_MQ(h,w);

    /* update ranges */
    ac_h=ac_h2;
    ac_w=ac_w2;
    ac_w2<<=1;
    ac_h2 <<=1;
  }

  /* close arithmetic coder */
  ac_model_done(&acm_type);
  if (snr_image->root_max)
    ac_model_done(&acm_root);
  ac_model_done(&acm_vz);
  if (snr_image->valnz_max)
    ac_model_done(&acm_vnz);
  ac_model_done(&acm_res);
  ac_model_done(&acm_sign);
  ac_decoder_done(&acd);
}



/********************************************************
  Function Name
  -------------
  void clear_ZTR_D()

  Arguments
  ---------
  None.

  
  Description
  -----------
  Clear the zerotree descendent marks in the entire image.

  Functions Called
  ----------------
  None.


  Return Value
  ------------
  None.

********************************************************/ 
void clear_ZTR_D()
{
  register COEFFINFO **coeff;
  register i,j,dc_h2,dc_w2;

  coeff=coeffinfo;
  dc_h2=mzte_codec.dcHeight<<1;
  dc_w2=mzte_codec.dcWidth<<1;

  for(i=0;i<dc_h2;i++)
    for(j=dc_w2;j<width;j++)
      if(coeff[i][j].type == ZTR_D)
         coeff[i][j].type = UNTYPED;

  for(i=dc_h2;i<height;i++)
    for(j=0;j<width;j++)
      if(coeff[i][j].type == ZTR_D)
         coeff[i][j].type = UNTYPED;
}



/********************************************************
  Function Name
  -------------
  static void decode_pixel_MQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel in height and width
  
  Description
  -----------
  Decoding the type and/or value of a coefficient, a
  recursive function, multi quant mode.

  Functions Called
  ----------------
  ac_decode_symbol()
  mark_ZTR_D()
  mag_sign_decode_MQ()

  Return Value
  ------------
  None.

********************************************************/ 

static void decode_pixel_MQ(int h,int w)
{
  int zt_type;

#ifdef _CHECK_HI_LO_
  fprintf(acfp,"%3d, %3d: %7d %7d\n",h,w,acd.high,acd.low);
#endif


  /* zerotree descendent */
  if(coeffinfo[h][w].type==ZTR_D)
    return;

  /* decode zero tree symbol */
  if (coeffinfo[h][w].state == S_LEAF || coeffinfo[h][w].state == S_RLEAF
      || (coeffinfo[h][w].state == S_RVAL && coeffinfo[h][w].skip==0))
  {
    zt_type=VAL; /* tmp assign. for next switch statement */
  }
  else
  {
#ifdef _SHAPE_
    if(coeffinfo[h][w].mask==1) 
#endif
      
      zt_type=coeffinfo[h][w].type=ac_decode_symbol(&acd,&acm_type);
  
#ifdef _SHAPE_
    else { /* treat out-node as isolated zero for decoding purpose */
      zt_type = coeffinfo[h][w].type = IZ;
    }
  
#endif
  }
  /* mark ztr_d and decode magnitudes */
  switch(zt_type)
  {
    case ZTR:
      mark_ZTR_D(h,w);
    case IZ:
      coeffinfo[h][w].quantized_value=0;
      return;
    case VZTR:
      mark_ZTR_D(h,w);
    case VAL:
#ifdef _SHAPE_
      if(coeffinfo[h][w].mask==1) 
#endif
	mag_sign_decode_MQ(h,w);
      break;
    default:
      errorHandler("Invalid type in multi quant decoding.");     
  }


#ifdef _WRITE_COEFF_MAP_
  if (coeffinfo[h][w].state == S_LEAF || coeffinfo[h][w].state == S_RLEAF)
    coeffinfo[h][w].type = (coeffinfo[h][w].quantized_value) ? VLEAF : ZLEAF;
  else if (coeffinfo[h][w].state == S_RVAL && coeffinfo[h][w].skip==0) 
    coeffinfo[h][w].type = VAL;
#endif _WRITE_COEFF_MAP_

}


/********************************************************
  Function Name
  -------------
  static void mark_ZTR_D(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  
  Description
  -----------
  Mark the coefficient at (h,w) and its descendents as
  zerotree descendents. 

  Functions Called
  ----------------
  mark_ZTR_D()


  Return Value
  ------------
  None.

********************************************************/ 
static void mark_ZTR_D(int h,int w)
{
  int i,j;

  i=h<<1; j=w<<1;

  if(i<height && j<width)
  {
    coeffinfo[i][j].quantized_value     = 0;
    coeffinfo[i+1][j].quantized_value   = 0;
    coeffinfo[i][j+1].quantized_value   = 0;
    coeffinfo[i+1][j+1].quantized_value = 0;

    coeffinfo[i][j].type     = ZTR_D; 
    coeffinfo[i+1][j].type   = ZTR_D; 
    coeffinfo[i][j+1].type   = ZTR_D; 
    coeffinfo[i+1][j+1].type = ZTR_D; 

    mark_ZTR_D(i,j);
    mark_ZTR_D(i+1,j);
    mark_ZTR_D(i,j+1);
    mark_ZTR_D(i+1,j+1);
  }
}




/********************************************************
  Function Name
  -------------
  static void  mag_sign_decode_MQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  Description
  -----------
  Decode the value of a coefficient.

  Functions Called
  ----------------
  ac_decode_symbol()

  Return Value
  ------------
  None.

********************************************************/ 
static void mag_sign_decode_MQ(int h,int w)
{
  int val,v_sign,ac_model;

  if(coeffinfo[h][w].skip)
  {
    coeffinfo[h][w].quantized_value=0;
    return;
  }

  ac_model=stateToProbModel[coeffinfo[h][w].state];

  /* decode magnitude */
  switch(ac_model){
    case ACM_ROOT:  /* smallest three AC bands */
      val=ac_decode_symbol(&acd,&acm_root)+1;
      break;
    case ACM_VALNZ:
      val=ac_decode_symbol(&acd,&acm_vnz)+1;
      break;
    case ACM_VALZ:     /* leaf */
      val=ac_decode_symbol(&acd,&acm_vz);
      if(val ==0){
        coeffinfo[h][w].quantized_value=0;
        return;
      }
      else
        break;
    case ACM_RESID:  /* refinement, no sign */
      coeffinfo[h][w].quantized_value=ac_decode_symbol(&acd,&acm_res);
      return;
    default:
      errorHandler("not a valid magnitude type in encode_pixel_MQ");
  }

  v_sign=ac_decode_symbol(&acd,&acm_sign);
  coeffinfo[h][w].quantized_value=(v_sign ==0) ? val: -val;
}



/********************************************************
  Function Name
  -------------
  void write_coeff_to_file(char *fname,WVT_CODEC *mzte_codec)

  Arguments
  ---------
  char *fname - output file name
  WVT_CODEC *mzte_codec - structure for the mzte coder.

  
  Description
  -----------
  write coeff position, value, type, state, and ac_model to
  a file in the band-wise order.

  Functions Called
  ----------------
  None.


  Return Value
  ------------
  None.

********************************************************/ 
void write_coeff_to_file(char *fname,WVT_CODEC *mzte_codec)
{
  int h,w,ac_h,ac_w,ac_h2,ac_w2;
  FILE *fp;
  COEFFINFO **coeffinfo;

  coeffinfo=mzte_codec->SPlayer[color].coeffinfo;
  fp=fopen(fname,"w");

  ac_h=mzte_codec->dcHeight;
  ac_w=mzte_codec->dcWidth;
  ac_h2=ac_h<<1; ac_w2=ac_w<<1;
  for(h=0;h<ac_h;h++)
    for(w=0;w<ac_w;w++)
      fprintf(fp,"%4d,%4d: %6d %3d %3d\n",h,w,
        coeffinfo[h][w].quantized_value,coeffinfo[h][w].type,
        coeffinfo[h][w].state);

  while(ac_h2<=mzte_codec->Image[0].height && 
        ac_w2<=mzte_codec->Image[0].width){
    /* LH */
    for(h=0;h<ac_h;h++)
      for(w=ac_w;w<ac_w2;w++)
      fprintf(fp,"%4d,%4d: %6d %3d %3d\n",h,w,
        coeffinfo[h][w].quantized_value,coeffinfo[h][w].type,
        coeffinfo[h][w].state);

    /* HL */
    for(h=ac_h;h<ac_h2;h++)
      for(w=0;w<ac_w;w++)
      fprintf(fp,"%4d,%4d: %6d %3d %3d\n",h,w,
        coeffinfo[h][w].quantized_value,coeffinfo[h][w].type,
        coeffinfo[h][w].state);

    /* HH */
    for(h=ac_h;h<ac_h2;h++)
      for(w=ac_w;w<ac_w2;w++)
      fprintf(fp,"%4d,%4d: %6d %3d %3d\n",h,w,
        coeffinfo[h][w].quantized_value,coeffinfo[h][w].type,
        coeffinfo[h][w].state);

    /* update ranges */
    ac_h=ac_h2;
    ac_w=ac_w2;
    ac_w2<<=1;
    ac_h2 <<=1;
  }
  fclose(fp);
}

