/* $Id: ztscan_enc.c,v 1.41 1998/02/20 21:14:40 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

/************************************************************/
/*  Filename: ztscan_enc.c                                  */
/*  Author: Bing-Bing Chai                                  */
/*  Date: Dec. 4, 1997                                      */
/*                                                          */
/*  Descriptions:                                           */
/*    This file contains the routines that performs         */
/*    zero tree scanning and entropy encoding.              */
/*                                                          */
/************************************************************/

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <ctype.h>
#include <string.h>
#include <math.h>

#include "startcode.h"
#include "dataStruct.h"
#include "states.h"
#include "globals.h"
#include "errorHandler.h"

#include "ac.h"
#include "bitpack.h"
#include "msg.h"

/* definations related to arithmetic coding */
#define NUMCHAR_TYPE 4
#define ADAPT 1
#define PROB_MOD_ORDER AC_ORDER_ZERO
#define NO_CONTEXT -1
#define valSlotMax 10
#define valInc 15

/* local global variables */
static color=0,bit_stream_length;
static unsigned char *bitbuffer=NULL;
static COEFFINFO **coeffinfo;
static DATA **dc_coeff;


/* arithmetic models related variables, refer to syntax for meaning */
static ac_encoder ace;
static ac_model acm_type, acm_vz, acm_vnz, acm_res, acm_root,acm_sign;

extern void clear_ZTR_D();


/* functions defined in this file */
void wavelet_dc_encode(int c);
static void cacll_encode();
void wavelet_higher_bands_encode_SQ(int c);
static void cachb_encode_SQ(SNR_IMAGE *snr_img);
static void encode_pixel_SQ(int h,int w);
static void  mag_sign_encode_SQ(int h,int w);
void wavelet_higher_bands_encode_MQ(SNR_IMAGE *snr_image, int c);
static void cachb_encode_MQ(SNR_IMAGE *snr_image);
static void encode_pixel_MQ(int h,int w);
static void  mag_sign_encode_MQ(int h,int w);
static void mark_ZTR_D(int h,int w);


static void ERROR(char *msg)
{
  printf("%s\n",msg);
  exit(-1);
} 



static void free_bitbuffer()
{
  free(bitbuffer);
  bitbuffer=NULL;
}




/* The following functions assume that mzte_codec is a global variable */
extern WVT_CODEC mzte_codec;

#ifdef _CHECK_HI_LO_
extern FILE *acfp;
#endif


/* This program combine the stuff in bitbuffer into the actual bitstream */
/* This function replaces the two states following each ac_encoder done */
static void write_to_bitstream(unsigned char *bitbuffer,int total_bits)
{
  int i,bit_stream_length=total_bits>>3, resid=total_bits%8;

  /* write to file assume file has been opened by control program*/
  for(i=0;i<bit_stream_length;i++)
    emit_bits(bitbuffer[i],8);

  if(resid != 0)
    emit_bits(bitbuffer[bit_stream_length]>>(8-resid),resid);
}


/******************************************************************/
/****************************  DC  ********************************/
/******************************************************************/

/*******************************************************/
/**************  Forward DC Prediction  ****************/
/*******************************************************/

/********************************************************
  Function Name
  -------------
  static DATA DC_pred_pix(int i, int j)

  Arguments
  ---------
  int i, int j: Index of wavelet coefficient (row, col)
  
  Description
  -----------
  DPCM prediction for a DC coefficient, refer to syntax
  for algorithm. 

  Functions Called
  ----------------
  None.

  Return Value
  ------------
    prediction for coeffinfo[i][j].quantized_value
********************************************************/ 
static DATA DC_pred_pix(int i, int j)
{
  if(i==0){
    if(j==0)
      return(0);
    else
      return(dc_coeff[i][j-1]);
  }
  else if(j==0)
    return(dc_coeff[i-1][j]);

  if(abs(dc_coeff[i-1][j-1]-dc_coeff[i][j-1])<
       abs(dc_coeff[i-1][j-1]-dc_coeff[i-1][j]))
    return(dc_coeff[i-1][j]);
  else
    return(dc_coeff[i][j-1]);
}



/*****************************************************
  Function Name
  -------------
  void DC_predict()

  Arguments
  ---------
  None
  
  Description
  -----------
  control program for DC prediction

  Functions Called
  ----------------
  DC_pred_pix(i,j).

  Return Value
  ------------
  none
******************************************************/

static void DC_predict(int color)
{
  int i,j,dc_h,dc_w,offset_dc,max_dc;

  dc_h=mzte_codec.dcHeight;
  dc_w=mzte_codec.dcWidth;

  dc_coeff=(DATA **)calloc(dc_h,sizeof(DATA *));
  for(i=0;i<dc_h;i++)
    dc_coeff[i]=(DATA *)calloc(dc_w,sizeof(DATA));

  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;

  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++)
      dc_coeff[i][j]=coeffinfo[i][j].quantized_value;

  /* prediction */    
  offset_dc=coeffinfo[0][0].quantized_value;
  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++){
      if(offset_dc>(coeffinfo[i][j].quantized_value-=DC_pred_pix(i,j)))
        offset_dc=coeffinfo[i][j].quantized_value;
    }

  if(offset_dc>0)
    offset_dc=0;

  /* adjust coeff's by offset_dc */
  max_dc=coeffinfo[0][0].quantized_value-offset_dc;
  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++){
      coeffinfo[i][j].quantized_value -=offset_dc;
      /* find max_dc */
      if(coeffinfo[i][j].quantized_value>max_dc)
        max_dc=coeffinfo[i][j].quantized_value;
    }
    
  mzte_codec.offset_dc=offset_dc;
  mzte_codec.max_dc=max_dc;  /* hjlee */
  noteDebug("DC pred: offset=%d, max_dc=%d",
	    mzte_codec.offset_dc,mzte_codec.max_dc);  

  for(i=0;i<dc_h;i++)
    free(dc_coeff[i]);
  free(dc_coeff);
}




/********************************************************
  Function Name
  -------------
  void wavelet_dc_encode(int c)


  Arguments
  ---------
  int c - color component.
  
  Description
  -----------
  Control program for encode DC information for one 
  color component.

  Functions Called
  ----------------
  None.
  DC_predict()
  put_param()
  cacll_encode()
  
  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_dc_encode(int c)
{

  noteDetail("Encoding DC (wavelet_dc_encode)....");
  color=c;

  emit_bits(mzte_codec.mean[color], 8);
  put_param(mzte_codec.Qdc[color], 7);
  /* emit_bits(mzte_codec.Qdc[color], 8); */
  DC_predict(color);
  put_param(-mzte_codec.offset_dc,7);
  /* put_param(mzte_codec.max_dc,7); */
  put_param(mzte_codec.max_dc+mzte_codec.offset_dc,7);  /* hjlee */
  
#ifdef _CHECK_HI_LO_
  acfp=fopen("encode_hl.dat","w");
#endif

  cacll_encode();
  noteDetail("Completed encoding DC.");

#ifdef _CHECK_HI_LO_
  fclose(acfp);
#endif

}



/********************************************************
  Function Name
  -------------
  static void cacll_encode()

  Arguments
  ---------
  None.

  
  Description
  -----------
  Encode DC information for one color component.

  Functions Called
  ----------------
  ac_encoder_init()
  ac_model_init()
  ac_encode_symbol()
  ac_model_done()
  ac_encoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void cacll_encode()
{
  int dc_h, dc_w,i,j,offset=mzte_codec.offset_dc;

  dc_w=mzte_codec.dcWidth;
  dc_h=mzte_codec.dcHeight;

  /* init arithmetic coder */
  ac_encoder_init(&ace,NULL,bitbuffer);
  ac_model_init(&acm_vz,mzte_codec.max_dc+1,
               NULL,ADAPT,PROB_MOD_ORDER,NO_CONTEXT,
               MIN(mzte_codec.max_dc+1,
                   valSlotMax),valInc);

  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;
  for(i=0;i<dc_h;i++)
    for(j=0;j<dc_w;j++){

#ifdef _CHECK_HI_LO_
      fprintf(acfp,"%3d, %3d: %7d %7d\n",i,j,ace.high,ace.low);
#endif

      ac_encode_symbol(&ace,&acm_vz,
                   coeffinfo[i][j].quantized_value);
    }

  /* close arithmetic coder */
  ac_model_done(&acm_vz);
  bit_stream_length=ac_encoder_done(&ace,&bitbuffer);

  /* write to file assume file has been opened by control program*/
  write_to_bitstream(bitbuffer,bit_stream_length);

  free_bitbuffer();
}



/*********************************************************************/
/*****************************  AC  **********************************/
/*************************  Single quant  ****************************/
/*********************************************************************/

/********************************************************
  Function Name
  -------------
  void wavelet_higher_bands_encode_SQ(int c)

  Arguments
  ---------
  int c - color component.
  
  Description
  -----------
  Control program for encoding AC information for one 
  color component. Single quant mode.

  Functions Called
  ----------------
  put_param()
  cachb_encode_SQ()

  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_higher_bands_encode_SQ(int c)
{
  SNR_IMAGE *snr_image;

  noteDetail("Encoding AC (wavelet_higher_bands_encode_SQ)....");
  color =c;
  snr_image=&(mzte_codec.SPlayer[color].SNRlayer.snr_image);

  /* header info */
  put_param(snr_image->root_max,7);
  emit_bits(MARKER_BIT, 1);
  put_param(snr_image->valz_max,7);
  put_param(snr_image->valnz_max,7);

  cachb_encode_SQ(snr_image);
  noteDetail("Completed encoding AC.");
}


static height,width;

/********************************************************
  Function Name
  -------------
  static void cachb_encode_SQ(SNR_IMAGE *snr_img)

  Arguments
  ---------
  SNR_IMAGE *snr_img - structure containing one color
                       component of an SNR image
  Description
  -----------
  Encode AC information for one color component. 
  Single quant mode, tree-depth scan

  Functions Called
  ----------------
  ac_encoder_init()
  ac_model_init()
  encode_pixel_SQ()
  ac_model_done()
  ac_encoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void cachb_encode_SQ(SNR_IMAGE *snr_img)
{
  int h,w,dc_h,dc_w,dc_h2,dc_w2;

  dc_h=mzte_codec.dcHeight;
  dc_w=mzte_codec.dcWidth;
  dc_h2=dc_h<<1;
  dc_w2=dc_w<<1;
  height=mzte_codec.Image[color].height;
  width=mzte_codec.Image[color].width;

  /* init arithmetic coder */
  ac_encoder_init(&ace,NULL,bitbuffer);
  ac_model_init(&acm_type,NUMCHAR_TYPE,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(NUMCHAR_TYPE,valSlotMax),valInc);
  if (snr_img->root_max)
    ac_model_init(&acm_root,snr_img->root_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_img->root_max,valSlotMax),valInc);
  ac_model_init(&acm_vz,snr_img->valz_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_img->valz_max+1,valSlotMax),valInc);
  if (snr_img->valnz_max)
    ac_model_init(&acm_vnz,snr_img->valnz_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_img->valnz_max,valSlotMax),valInc);
  ac_model_init(&acm_sign,2,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(2,valSlotMax),valInc);

  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;

  
  for(h=0;h<dc_h;h++)
    for(w=dc_w;w<dc_w2;w++){
      encode_pixel_SQ(h,w);           /* LH */
      encode_pixel_SQ(h+dc_h,w-dc_w); /* HL */
      encode_pixel_SQ(h+dc_h,w);      /* HH */
    }

  
  /* close arithmetic coder */
  ac_model_done(&acm_type);
  if (snr_img->root_max)
    ac_model_done(&acm_root);
  ac_model_done(&acm_vz);
  if (snr_img->valnz_max)
    ac_model_done(&acm_vnz);
  ac_model_done(&acm_sign);
  bit_stream_length=ac_encoder_done(&ace,&bitbuffer);

  /* write to file assume file has been opened by control program*/
  write_to_bitstream(bitbuffer,bit_stream_length);

  free_bitbuffer();
}



/********************************************************
  Function Name
  -------------
  static void encode_pixel_SQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel in height and width
  
  Description
  -----------
  Encoding the type and/or value of a coefficient, a
  recursive function.

  Functions Called
  ----------------
  mag_sign_encode_SQ()
  ac_encode_symbol()
  encode_pixel_SQ()

  Return Value
  ------------
  None.

********************************************************/ 
static void encode_pixel_SQ(int h,int w)
{
  int i,j;
  UNIT8 zt_type;


  /* code leave coefficients, value only, no type */
  if(coeffinfo[h][w].state == S_LEAF){
#ifdef _SHAPE_ /* skip out-node */
    if(coeffinfo[h][w].mask == 1) 
#endif
      mag_sign_encode_SQ(h,w);
    return;
  }
 
  /* code zerotree symbol */
#ifdef _SHAPE_ /* skip out-node */
  if(coeffinfo[h][w].mask == 1) 
#endif
   ac_encode_symbol(&ace,&acm_type,zt_type=coeffinfo[h][w].type);
#ifdef _SHAPE_
  else
    zt_type=coeffinfo[h][w].type;
#endif
  /* code magnitude and sign */
  /* For Arbitrary-Shape, out-node will always has zero coefficient,
     so only IZ or ZTR may be the zt_type -- SL*/

  switch(zt_type){
    case IZ : 
      break; /* will code the four children */
    case VZTR:
      mag_sign_encode_SQ(h,w);
    case ZTR:
#ifdef _SHAPE_
      if(coeffinfo[h][w].mask != 1) /* TBE for four children of out-node */ 
	break;
      else
#endif
     {
#ifdef _WRITE_COEFF_MAP_
      mark_ZTR_D(h,w);  /* not necessary, only for checking purpose */
#endif _WRITE_COEFF_MAP_
      return;
      }
    case VAL:
      mag_sign_encode_SQ(h,w);
      break;
    default: 
      errorHandler("invalid zerotree symbol in single quant encode");
  }

  /* scan children */
  i=h<<1; j=w<<1;
  encode_pixel_SQ(i,j);
  encode_pixel_SQ(i,j+1);
  encode_pixel_SQ(i+1,j);
  encode_pixel_SQ(i+1,j+1);
}


     
/********************************************************
  Function Name
  -------------
  static void  mag_sign_encode_SQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  Description
  -----------
  Encode the value of a coefficient.

  Functions Called
  ----------------
  ac_encode_symbol()

  Return Value
  ------------
  None.

********************************************************/ 
static void  mag_sign_encode_SQ(int h,int w)
{
  int val,v_sign;

  if((val=coeffinfo[h][w].quantized_value)>0)
    v_sign=0;
  else{
    val=-val;
    v_sign=1;
  }
  
  switch(stateToProbModel[coeffinfo[h][w].state]){    
    case ACM_ROOT:
      ac_encode_symbol(&ace,&acm_root,val-1);
      break;
    case ACM_VALNZ:
      ac_encode_symbol(&ace,&acm_vnz,val-1);
      break;
    case ACM_VALZ:
      ac_encode_symbol(&ace,&acm_vz,val);
      if(val==0)
        return;
      break;
    default:
      errorHandler("not a valid magnitude type in encode_pixel_SQ");
  }
  ac_encode_symbol(&ace,&acm_sign,v_sign);
}



/*********************************************************************/
/******************************  AC  *********************************/
/**************************  Multi quant  ****************************/
/*********************************************************************/

/********************************************************
  Function Name
  -------------
  void wavelet_higher_bands_encode_MQ(SNR_IMAGE *snr_image,int c)

  Arguments
  ---------
  SNR_IMAGE *snr_image - structure containing one color
                         component of an SNR image
  int c - color component.
  
  Description
  -----------
  Control program for encoding AC information for one 
  color component. Multi quant mode.

  Functions Called
  ----------------
  put_param()
  cachb_encode_MQ()

  Return Value
  ------------
  None.

********************************************************/ 
void wavelet_higher_bands_encode_MQ(SNR_IMAGE *snr_image,int c)
{
  noteDetail("Encoding AC (wavelet_higher_bands_encode_MQ)....");
  /* header info */
  put_param(snr_image->root_max,7);
  emit_bits(MARKER_BIT, 1);
  put_param(snr_image->valz_max,7);
  put_param(snr_image->valnz_max,7);
  color=c;
  coeffinfo=mzte_codec.SPlayer[color].coeffinfo;

  cachb_encode_MQ(snr_image);
}



/********************************************************
  Function Name
  -------------
  static void mark_ZTR_D(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  
  Description
  -----------
  Mark the coefficient at (h,w) and its descendents as
  zerotree descendents. 

  Functions Called
  ----------------
  mark_ZTR_D()


  Return Value
  ------------
  None.

********************************************************/ 
static void mark_ZTR_D(int h,int w)
{
  int i,j;

  i=h<<1; j=w<<1;

  if(i<height && j<width){
    coeffinfo[i][j].type     = ZTR_D;
    coeffinfo[i+1][j].type   = ZTR_D; 
    coeffinfo[i][j+1].type   = ZTR_D; 
    coeffinfo[i+1][j+1].type = ZTR_D; 
    mark_ZTR_D(i,j);
    mark_ZTR_D(i+1,j);
    mark_ZTR_D(i,j+1);
    mark_ZTR_D(i+1,j+1);
  }
}



/********************************************************
  Function Name
  -------------
  static void cachb_encode_MQ(SNR_IMAGE *snr_img)

  Arguments
  ---------
  SNR_IMAGE *snr_img - structure containing one color
                       component of an SNR image 
  Description
  -----------
  Encode AC information for one color component. 
  Multiple quant, bandwise scan.

  Functions Called
  ----------------
  ac_encoder_init()
  ac_model_init()
  clear_ZTR_D();
  encode_pixel_MQ()
  ac_model_done()
  ac_encoder_done()

  Return Value
  ------------
  None.

********************************************************/ 
static void cachb_encode_MQ(SNR_IMAGE *snr_image)
{
  int h,w,ac_h,ac_w,ac_h2,ac_w2;

  /* ac_h, ac_w init */
  ac_h=mzte_codec.dcHeight;
  ac_w=mzte_codec.dcWidth;
  ac_h2=ac_h<<1;
  ac_w2=ac_w<<1;
  height=mzte_codec.SPlayer[color].height;
  width=mzte_codec.SPlayer[color].width;

  /* init arithmetic coder */
  ac_encoder_init(&ace,NULL,bitbuffer);
  ac_model_init(&acm_type,NUMCHAR_TYPE,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(NUMCHAR_TYPE,valSlotMax),valInc);
  if (snr_image->root_max)
    ac_model_init(&acm_root,snr_image->root_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_image->root_max,valSlotMax),valInc);
  ac_model_init(&acm_vz,snr_image->valz_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_image->valz_max+1,valSlotMax),valInc);
  if (snr_image->valnz_max)
    ac_model_init(&acm_vnz,snr_image->valnz_max,NULL,ADAPT,PROB_MOD_ORDER,
		  NO_CONTEXT,MIN(snr_image->valnz_max,valSlotMax),valInc);
  ac_model_init(&acm_res,snr_image->residual_max+1,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(snr_image->residual_max+1,valSlotMax),valInc);
  ac_model_init(&acm_sign,2,NULL,ADAPT,PROB_MOD_ORDER,
                NO_CONTEXT,MIN(2,valSlotMax),valInc);


  /* clear the ZTR_D type from the previous pass */
  clear_ZTR_D();


  /* scan each coefficients in the spatial layer */
  while(ac_h2<=height && ac_w2<=width){
    /* LH */
    for(h=0;h<ac_h;h++)
      for(w=ac_w;w<ac_w2;w++)
        encode_pixel_MQ(h,w);

    /* HL */
    for(h=ac_h;h<ac_h2;h++)
      for(w=0;w<ac_w;w++)
        encode_pixel_MQ(h,w);

    /* HH */
    for(h=ac_h;h<ac_h2;h++)
      for(w=ac_w;w<ac_w2;w++)
        encode_pixel_MQ(h,w);

    /* update ranges */
    ac_h=ac_h2;
    ac_w=ac_w2;
    ac_w2<<=1;
    ac_h2 <<=1;
  }

  /* close arithmetic coder */
  ac_model_done(&acm_type);
  if (snr_image->root_max)
    ac_model_done(&acm_root);
  ac_model_done(&acm_vz);
  if (snr_image->valnz_max)
    ac_model_done(&acm_vnz);
  ac_model_done(&acm_res);
  ac_model_done(&acm_sign);
  bit_stream_length=ac_encoder_done(&ace,&bitbuffer);


  /* write to file assume file has been opened by control program*/
  write_to_bitstream(bitbuffer,bit_stream_length);

  free_bitbuffer();
}




/********************************************************
  Function Name
  -------------
  static void encode_pixel_MQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel in height and width
  
  Description
  -----------
  Encoding the type and/or value of a coefficient, a
  recursive function, multi quant mode.

  Functions Called
  ----------------
  ac_encode_symbol()
  mark_ZTR_D()
  mag_sign_encode_MQ()
 
  Return Value
  ------------
  None.

********************************************************/ 
static void encode_pixel_MQ(int h,int w)
{
  int zt_type;

#ifdef _CHECK_HI_LO_
  fprintf(acfp,"%3d, %3d: %7d %7d\n",h,w,ace.high,ace.low);
#endif

  /* zerotree descendent */
  if(coeffinfo[h][w].type==ZTR_D)
    return;

  /* encode zero tree symbol */
  
  if (coeffinfo[h][w].state == S_LEAF || coeffinfo[h][w].state == S_RLEAF
      || (coeffinfo[h][w].state == S_RVAL && coeffinfo[h][w].skip==0))
  {
    zt_type = VAL;

#ifdef _WRITE_COEFF_MAP_
    if (coeffinfo[h][w].state == S_RVAL && coeffinfo[h][w].skip==0) 
      coeffinfo[h][w].type = VAL;
#endif _WRITE_COEFF_MAP_
  }

  else
  {
    zt_type=coeffinfo[h][w].type;
#ifdef _SHAPE_
    if(coeffinfo[h][w].mask==1) /* skip out-node */  
#endif
      ac_encode_symbol(&ace,&acm_type,zt_type);
  }


  /* mark ztr_d and encode magnitudes */
  switch(zt_type){
    case ZTR:
#ifdef _SHAPE_
      if(coeffinfo[h][w].mask==1) /* mark ZTR-D for in-node root */
#endif
      mark_ZTR_D(h,w);
    case IZ:
      return;
    case VZTR:
      mark_ZTR_D(h,w);
    case VAL:
    case VLEAF:
    case ZLEAF:
#ifdef _SHAPE_
      if(coeffinfo[h][w].mask==1) /* only code for in-node*/
#endif
      mag_sign_encode_MQ(h,w);
      break;
    default:
      errorHandler("Invalid type in multi quant decoding.");     
  }

}




        
/********************************************************
  Function Name
  -------------
  static void  mag_sign_encode_MQ(int h,int w)

  Arguments
  ---------
  int h,int w - position of a pixel

  Description
  -----------
  Encode the value of a coefficient.

  Functions Called
  ----------------
  ac_encode_symbol()

  Return Value
  ------------
  None.

********************************************************/ 
 static void  mag_sign_encode_MQ(int h,int w)
{
  int val,v_sign,ac_model;

  if(coeffinfo[h][w].skip)
    return;

  ac_model=stateToProbModel[coeffinfo[h][w].state];

  if((val=coeffinfo[h][w].quantized_value)>=0)
    v_sign=0;
  else{
    val=-val;
    v_sign=1;
  }
  
  /* code magnitude */
  switch(ac_model){
    case ACM_ROOT:  /* smallest three AC bands */
      ac_encode_symbol(&ace,&acm_root,val-1);
      break;
    case ACM_VALNZ:
      ac_encode_symbol(&ace,&acm_vnz,val-1);
      break;
    case ACM_VALZ:     /* leaf */
      if(val == 0){
        ac_encode_symbol(&ace,&acm_vz,0);
        return;
      }
      else
        ac_encode_symbol(&ace,&acm_vz,val);
      break;
    case ACM_RESID:  /* refinement, no sign */
      ac_encode_symbol(&ace,&acm_res,val);
      return;
    default:
      errorHandler("In coefficient (%d,%d). Invalid model (=%d) in \
encode_pixel_MQ.",h,w,ac_model);
  }
  
  ac_encode_symbol(&ace,&acm_sign,v_sign);
}







